import subprocess

import attr
import ui

import tbump


@attr.s
class Hook:
    name = attr.ib()
    cmd = attr.ib()


class HookError(tbump.Error):
    def print_error(self):
        ui.error(
            ui.reset, "`%s`" % self.cmd,
            "exited with return code", self.rc
        )


def print_hook(hook):
    ui.info(ui.darkgray, "$", ui.reset, hook.cmd)


class HooksRunner:
    def __init__(self, working_path=None):
        self.hooks = list()
        self.working_path = working_path

    def add_hook(self, hook):
        self.hooks.append(hook)

    def run(self, new_version):
        for i, hook in enumerate(self.hooks):
            hook.cmd = hook.cmd.format(new_version=new_version)
            ui.info_count(i, len(self.hooks), ui.bold, hook.name)
            print_hook(hook)
            rc = subprocess.call(hook.cmd, shell=True, cwd=self.working_path)
            if rc != 0:
                raise HookError(name=hook.name, cmd=hook.cmd, rc=rc)



class Error(Exception):
    def __init__(self, **kwargs):
        self._dict = kwargs
        for k, v in kwargs.items():
            setattr(self, k, v)

    def print_error(self):
        pass

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, repr(self._dict))



import os

import path
import pytest

import tbump.git

from ui.tests.conftest import message_recorder
message_recorder  # silence pyflakes


@pytest.fixture()
def tmp_path(tmpdir):
    return path.Path(tmpdir)


@pytest.fixture
def test_data_path():
    this_dir = path.Path(__file__).abspath().parent
    return this_dir.joinpath("data")


@pytest.fixture(autouse=True, scope="session")
def restore_cwd():
    old_cwd = os.getcwd()
    yield
    os.chdir(old_cwd)


def assert_in_file(*args):
    parts = args[0:-1]
    first = parts[0]
    rest = parts[1:]
    last = parts[-1]
    expected_line = args[-1]
    file_path = path.Path(first).joinpath(*rest)
    for line in file_path.lines():
        if expected_line in line:
            return
    assert False, "No line found matching %s in %s" % (expected_line, last)


def setup_repo(tmp_path, test_data_path):
    src_path = tmp_path.joinpath("src")
    test_data_path.copytree(src_path)
    tbump.git.run_git(src_path, "init")
    tbump.git.run_git(src_path, "add", ".")
    tbump.git.run_git(src_path, "commit", "--message", "initial commit")
    tbump.git.run_git(src_path, "tag",
                      "--annotate",
                      "--message", "v1.2.41-alpha-1",
                      "v1.2.41-alpha-1")
    return src_path


def setup_remote(tmp_path):
    git_path = tmp_path.joinpath("git")
    git_path.mkdir()
    remote_path = git_path.joinpath("repo.git")
    remote_path.mkdir()
    tbump.git.run_git(remote_path, "init", "--bare")

    src_path = tmp_path.joinpath("src")
    tbump.git.run_git(src_path, "remote", "add", "origin", remote_path)
    tbump.git.run_git(src_path, "push", "-u", "origin", "master")
    return src_path


@pytest.fixture
def test_repo(tmp_path, test_data_path):
    res = setup_repo(tmp_path, test_data_path)
    setup_remote(tmp_path)
    return res



import toml
import pytest

import tbump.git
import tbump.hooks
import tbump.main
from tbump.test.conftest import assert_in_file


@pytest.fixture
def working_hook(test_repo):
    return add_hook(test_repo, "fake yarn", "python yarn.py")


@pytest.fixture
def crashing_hook(test_repo):
    return add_hook(test_repo, "crash", "python nosuchfile.py")


def add_hook(test_repo, name, cmd):
    """ Patch the configuration file so that we can also test hooks.

    """
    cfg_path = test_repo / "tbump.toml"
    parsed = toml.loads(cfg_path.text())
    if "hook" not in parsed:
        parsed["hook"] = list()
    parsed["hook"].append({"cmd": cmd, "name": name})
    cfg_path.write_text(toml.dumps(parsed))
    tbump.git.run_git(test_repo, "add", ".")
    tbump.git.run_git(test_repo, "commit", "--message", "add yarn hook")


def test_end_to_end(test_repo, working_hook):
    tbump.main.main(["-C", test_repo, "1.2.41-alpha-2", "--non-interactive"])

    assert_in_file("yarn.lock", "1.2.41-alpha-2")


def test_hook_fails(test_repo, working_hook, crashing_hook):
    with pytest.raises(tbump.hooks.HookError):
        tbump.main.run(["-C", test_repo, "1.2.41-alpha-2", "--non-interactive"])



import abc
import argparse
import os
import sys

import path
import ui

import tbump
import tbump.config
import tbump.git
from tbump.file_bumper import FileBumper
from tbump.git_bumper import GitBumper
from tbump.hooks import HooksRunner


TBUMP_VERSION = "3.0.0"


class InvalidConfig(tbump.Error):
    def __init__(self, io_error=None, parse_error=None):
        self.io_error = io_error
        self.parse_error = parse_error

    def print_error(self):
        if self.io_error:
            ui.error("Could not read config file:", self.io_error)
        if self.parse_error:
            ui.error("Invalid config:",  self.parse_error)


class Cancelled(tbump.Error):
    def print_error(self):
        ui.error("Cancelled by user")


def main(args=None):
    # Supress backtrace if exception derives from tbump.Error
    try:
        run(args)
    except tbump.Error as error:
        error.print_error()
        sys.exit(1)


def parse_command_line(cmd):
    parser = argparse.ArgumentParser()
    parser.add_argument("new_version")
    parser.add_argument("-C", "--cwd", dest="working_dir")
    parser.add_argument("--non-interactive", dest="interactive", action="store_false")
    parser.add_argument("--version", action="version", version=TBUMP_VERSION)
    args = parser.parse_args(args=cmd)
    return args


class Runner(metaclass=abc.ABCMeta):
    def __init__(self, args):
        self.new_version = args.new_version

        working_dir = args.working_dir or os.getcwd()
        self.working_path = path.Path(working_dir)
        os.chdir(self.working_path)

        self.config = self.parse_config()

        self.git_bumper = self.setup_git_bumper()
        self.file_bumper = self.setup_file_bumper()
        self.hooks_runner = self.setup_hooks_runner()

    def display_bump(self, dry_run=False):
        ui.info_1(
            "Bumping from",
            ui.reset, ui.bold, self.config.current_version,
            ui.reset, "to",
            ui.reset, ui.bold, self.new_version, end=""
        )
        if dry_run:
            ui.info(ui.brown, "(dry _run)")
        else:
            ui.info()
        ui.info()

    def parse_config(self):
        tbump_path = self.working_path.joinpath("tbump.toml")
        try:
            config = tbump.config.parse(tbump_path)
        except IOError as io_error:
            raise InvalidConfig(io_error=io_error)
        except Exception as parse_error:
            raise InvalidConfig(parse_error=parse_error)
        return config

    def setup_git_bumper(self):
        git_bumper = GitBumper(self.working_path)
        git_bumper.set_config(self.config)
        return git_bumper

    def setup_file_bumper(self):
        file_bumper = FileBumper(self.working_path)
        file_bumper.set_config(self.config)
        return file_bumper

    def setup_hooks_runner(self):
        hooks_runner = HooksRunner(self.working_path)
        for hook in self.config.hooks:
            hooks_runner.add_hook(hook)
        return hooks_runner

    def before_bump(self, patches, git_commands):
        pass

    def bump(self):
        self.display_bump()
        patches = self.file_bumper.compute_patches(self.new_version)
        git_commands = self.git_bumper.compute_commands(self.new_version)
        self.before_bump(patches, git_commands)

        ui.info_2("Patching files")
        self.file_bumper.apply_patches(patches)
        if self.hooks_runner.hooks:
            ui.info()
            ui.info_2("Running hooks")
            self.hooks_runner.run(self.new_version)
        ui.info()
        ui.info_2("Running git commands", ui.ellipsis)
        self.git_bumper.run_commands(git_commands)
        ui.info()
        ui.info(ui.green, "Done", ui.check)


class InteractiveRunner(Runner):
    def __init__(self, args):
        super().__init__(args)

    def check(self):
        try:
            self.git_bumper.check_state(self.new_version)
        except tbump.git_bumper.NoTrackedBranch as e:
            e.print_error()
            proceed = ui.ask_yes_no("Continue anyway?", default=False)
            if not proceed:
                raise Cancelled from None

    def display_patches(self, patches):
        for patch in patches:
            tbump.file_bumper.print_patch(patch)

    def display_git_commands(self, git_commands):
        for git_command in git_commands:
            tbump.git.print_git_command(git_command)

    def display_hooks(self):
        hooks = self.hooks_runner.hooks
        if not hooks:
            return
        ui.info_2("Would execute these hooks")
        for i, hook in enumerate(hooks):
            ui.info_count(i, len(hooks), ui.bold, hook.name)
            tbump.hooks.print_hook(hook)

    def before_bump(self, patches, git_commands):
        ui.info_2("Would patch those files")
        self.display_patches(patches)
        ui.info()
        self.display_hooks()
        ui.info()
        ui.info_2("Would run these commands")
        self.display_git_commands(git_commands)
        ui.info()
        answer = ui.ask_yes_no("Looking good?", default=False)
        if not answer:
            raise Cancelled from None
        ui.info()


class NonInteractiveRunner(Runner):
    def check(self):
        self.git_bumper.check_state(self.new_version)
        return True


def run(cmd=None):
    args = parse_command_line(cmd)
    interactive = args.interactive
    if interactive:
        runner = InteractiveRunner(args)
    else:
        runner = NonInteractiveRunner(args)

    runner.check()
    runner.bump()



import subprocess

import ui

import tbump


class GitError(tbump.Error):
    pass


class GitCommandError(GitError):
    def print_error(self):
        cmd_str = "git " + " ".join(self.cmd)
        ui.error("Command", "`%s`" % cmd_str, "failed")


def print_git_command(cmd):
    ui.info(ui.darkgray, "$", ui.reset, "git", *cmd)


def run_git(working_path, *cmd, verbose=False):
    """ Run git `cmd` in given `working_path`

    Displays the command ran if `verbose` is True

    Raise GitCommandError if return code is non-zero.
    """
    if verbose:
        print_git_command(cmd)
    git_cmd = list(cmd)
    git_cmd.insert(0, "git")

    returncode = subprocess.call(git_cmd, cwd=working_path)
    if returncode != 0:
        raise GitCommandError(cmd=cmd, working_path=working_path)


def run_git_captured(working_path, *cmd, check=True):
    """ Run git `cmd` in given `working_path`, capturing the output

    Return a tuple (returncode, output).

    Raise GitCommandError if return code is non-zero and check is True
    """
    git_cmd = list(cmd)
    git_cmd.insert(0, "git")
    options = dict()
    options["stdout"] = subprocess.PIPE
    options["stderr"] = subprocess.STDOUT

    ui.debug(ui.lightgray, working_path, "$", ui.reset, *git_cmd)
    process = subprocess.Popen(git_cmd, cwd=working_path, **options)
    output, _ = process.communicate()
    output = output.decode("utf-8")
    if output.endswith('\n'):
        output = output.strip('\n')
    returncode = process.returncode
    ui.debug(ui.lightgray, "[%i]" % returncode, ui.reset, output)
    if check and returncode != 0:
        raise GitCommandError(working_path=working_path, cmd=cmd, output=output)
    return returncode, output



import attr
import path
import ui

import tbump
import tbump.config
import tbump.git


def print_patch(patch):
    ui.info(
        ui.red, "- ", ui.reset,
        ui.bold, patch.src, ":", ui.reset,
        ui.darkgray, patch.lineno + 1, ui.reset,
        " ", ui.red, patch.old_line.strip(),
        sep=""
    )
    ui.info(
        ui.green, "+ ", ui.reset,
        ui.bold, patch.src, ":", ui.reset,
        ui.darkgray, patch.lineno + 1, ui.reset,
        " ", ui.green, patch.new_line.strip(),
        sep=""
    )


@attr.s
class ChangeRequest:
    src = attr.ib()
    old_string = attr.ib()
    new_string = attr.ib()
    search = attr.ib(default=None)


@attr.s
class Patch:
    src = attr.ib()
    lineno = attr.ib()
    old_line = attr.ib()
    new_line = attr.ib()


class BadSubstitution(tbump.Error):
    def print_error(self):
        message = [
            " ", self.src + ":",
            " refusing to ", self.verb, " version containing 'None'\n",
        ]
        message += [
            "More info:\n",
            " * version groups:  ", self.groups, "\n"
            " * template:        ", self.template, "\n",
            " * version:         ", self.version, "\n",
        ]
        ui.error(*message, end="", sep="")


class InvalidVersion(tbump.Error):
    def print_error(self):
        ui.error("Could not parse", self.version, "as a valid version string")


class SourceFileNotFound(tbump.Error):
    def print_error(self):
        ui.error(self.src, "does not exist")


class CurrentVersionNotFound(tbump.Error):
    # TODO: raise just once for all errors
    def print_error(self):
        ui.error(
            "Current version string: (%s)" % self.current_version_string,
            "not found in", self.src
        )


def should_replace(line, old_string, search=None):
    if not search:
        return old_string in line
    else:
        return (old_string in line) and (search in line)


def on_version_containing_none(src, verb, version, *, groups, template):
    raise BadSubstitution(src=src, verb=verb, version=version, groups=groups, template=template)


class FileBumper():
    def __init__(self, working_path):
        self.working_path = working_path
        self.files = None
        self.version_regex = None
        self.current_version = None
        self.current_groups = None
        self.new_version = None
        self.new_groups = None

    def parse_version(self, version):
        match = self.version_regex.fullmatch(version)
        if match is None:
            raise InvalidVersion(version=version, regex=self.version_regex)
        return match.groupdict()

    def set_config(self, config):
        self.files = config.files
        self.check_files_exist()
        self.version_regex = config.version_regex
        self.current_version = config.current_version
        self.current_groups = self.parse_version(self.current_version)

    def check_files_exist(self):
        for file in self.files:
            expected_path = self.working_path.joinpath(file.src)
            if not expected_path.exists():
                raise SourceFileNotFound(src=file.src)

    def compute_patches(self, new_version):
        self.new_version = new_version
        self.new_groups = self.parse_version(self.new_version)
        change_requests = self.compute_change_requests()
        tbump_toml_change = ChangeRequest("tbump.toml", self.current_version, new_version)
        change_requests.append(tbump_toml_change)
        patches = list()
        for change_request in change_requests:
            patches_for_request = self.compute_patches_for_change_request(change_request)
            patches.extend(patches_for_request)
        return patches

    def compute_patches_for_change_request(self, change_request):
        old_string = change_request.old_string
        new_string = change_request.new_string
        search = change_request.search

        file_path = path.Path(self.working_path).joinpath(change_request.src)
        old_lines = file_path.lines()

        patches = list()
        for i, old_line in enumerate(old_lines):
            if should_replace(old_line, old_string, search):
                new_line = old_line.replace(old_string, new_string)
                patch = Patch(change_request.src, i, old_line, new_line)
                patches.append(patch)
        if not patches:
            raise CurrentVersionNotFound(src=change_request.src, current_version_string=old_string)
        return patches

    def compute_change_requests(self):
        change_requests = list()
        for file in self.files:
            change_request = self.compute_change_request_for_file(file)
            change_requests.append(change_request)
        return change_requests

    def compute_change_request_for_file(self, file):
        if file.version_template:
            current_version = file.version_template.format(**self.current_groups)
            if "None" in current_version:
                on_version_containing_none(
                    file.src,
                    "look for",
                    current_version,
                    groups=self.current_groups,
                    template=file.version_template
                )
            new_version = file.version_template.format(**self.new_groups)
            if "None" in new_version:
                on_version_containing_none(
                    file.src,
                    "replace by",
                    new_version,
                    groups=self.new_groups,
                    template=file.version_template
                )
        else:
            current_version = self.current_version
            new_version = self.new_version

        to_search = None
        if file.search:
            to_search = file.search.format(current_version=current_version)

        return ChangeRequest(file.src, current_version, new_version, search=to_search)

    def apply_patches(self, patches):
        for patch in patches:
            print_patch(patch)
            file_path = path.Path(self.working_path).joinpath(patch.src)
            # TODO: read and write each file only once?
            lines = file_path.lines()
            lines[patch.lineno] = patch.new_line
            file_path.write_lines(lines)



import ui

import tbump.git


class DirtyRepository(tbump.git.GitError):

    def print_error(self):
        ui.error("Repository is dirty")
        ui.info(self.git_status_output)


class NotOnAnyBranch(tbump.git.GitError):
    def print_error(self):
        ui.error("Not on any branch")


class NoTrackedBranch(tbump.git.GitError):
    def print_error(self):
        ui.error("Current branch (%s)" % self.branch,
                 "does not track anything. Cannot push.")


class RefAlreadyExists(tbump.git.GitError):
    def print_error(self):
        ui.error("git ref", self.ref, "already exists")


class GitBumper():
    def __init__(self, repo_path):
        self.repo_path = repo_path
        self.tag_template = None
        self.message_template = None
        self.remote_name = None
        self.remote_branch = None

    def set_config(self, config):
        self.tag_template = config.tag_template
        self.message_template = config.message_template

    def run_git(self, *args, **kwargs):
        full_args = [self.repo_path] + list(args)
        return tbump.git.run_git(*full_args, **kwargs)

    def run_git_captured(self, *args, **kwargs):
        full_args = [self.repo_path] + list(args)
        return tbump.git.run_git_captured(*full_args, **kwargs)

    def check_dirty(self):
        _, out = self.run_git_captured("status", "--porcelain")
        dirty = False
        for line in out.splitlines():
            # Ignore untracked files
            if not line.startswith("??"):
                dirty = True
        if dirty:
            raise DirtyRepository(git_status_output=out)

    def get_current_branch(self):
        cmd = ("rev-parse", "--abbrev-ref", "HEAD")
        _, out = self.run_git_captured(*cmd)
        if out == "HEAD":
            raise NotOnAnyBranch()
        return out

    def get_tracking_ref(self):
        branch_name = self.get_current_branch()
        rc, out = self.run_git_captured(
            "rev-parse", "--abbrev-ref", "--symbolic-full-name", "@{u}",
            check=False)
        if rc != 0:
            raise NoTrackedBranch(branch=branch_name)
        return out

    def check_ref_does_not_exists(self, tag_name):
        rc, _ = self.run_git_captured("rev-parse", tag_name, check=False)
        if rc == 0:
            raise RefAlreadyExists(ref=tag_name)

    def check_state(self, new_version):
        self.check_dirty()
        self.get_current_branch()
        tag_name = self.tag_template.format(new_version=new_version)
        self.check_ref_does_not_exists(tag_name)

        tracking_ref = self.get_tracking_ref()
        self.remote_name, self.remote_branch = tracking_ref.split("/", maxsplit=1)

    def compute_commands(self, new_version):
        res = list()
        res.append(["add", "--update"])
        commit_message = self.message_template.format(new_version=new_version)
        res.append(["commit", "--message", commit_message])
        tag_name = self.tag_template.format(new_version=new_version)
        tag_message = tag_name
        res.append(["tag", "--annotate", "--message", tag_message, tag_name])
        if self.remote_branch:
            res.append(["push", self.remote_name, self.remote_branch, tag_name])
        return res

    def run_commands(self, commands):
        for command in commands:
            self.run_git(*command, verbose=True)



import pytest

import tbump.file_bumper
from tbump.test.conftest import assert_in_file


def test_file_bumper_simple(test_repo):
    bumper = tbump.file_bumper.FileBumper(test_repo)
    config = tbump.config.parse(test_repo.joinpath("tbump.toml"))
    assert bumper.working_path == test_repo
    bumper.set_config(config)
    patches = bumper.compute_patches(new_version="1.2.41-alpha-2")
    bumper.apply_patches(patches)

    assert_in_file(test_repo, "package.json", '"version": "1.2.41-alpha-2"')
    assert_in_file(test_repo, "package.json", '"other-dep": "1.2.41-alpha-1"')
    assert_in_file(test_repo, "pub.js", "PUBLIC_VERSION = '1.2.41'")


def test_looking_for_empty_groups(tmp_path):
    tbump_path = tmp_path.joinpath("tbump.toml")
    tbump_path.write_text(
        r"""
        [version]
        current = "1.2"
        regex = '''
            (?P<major>\d+)
            \.
            (?P<minor>\d+)
            (
              \.
              (?P<patch>\d+)
            )?
        '''

        [git]
        message_template = "Bump to {new_version}"
        tag_template = "v{new_version}"

        [[file]]
        src = "foo"
        version_template = "{major}.{minor}.{patch}"

        """
    )
    foo_path = tmp_path.joinpath("foo")
    foo_path.write_text(
        """
        version = "1.2"
        """
    )
    config = tbump.config.parse(tbump_path)
    bumper = tbump.file_bumper.FileBumper(tmp_path)
    bumper.set_config(config)
    with pytest.raises(tbump.file_bumper.BadSubstitution) as e:
        bumper.compute_patches(new_version="1.3.1")
    assert e.value.src == "foo"
    assert e.value.groups == {"major": "1", "minor": "2", "patch": None}


def test_current_version_not_found(tmp_path):
    tbump_path = tmp_path.joinpath("tbump.toml")
    tbump_path.write_text(
        r"""
        [version]
        current = "1.2.3"
        regex = ".*"

        [git]
        message_template = "Bump to {new_version}"
        tag_template = "v{new_version}"

        [[file]]
        src = "version.txt"
        """
    )
    version_txt_path = tmp_path.joinpath("version.txt")
    version_txt_path.write_text("nope")
    config = tbump.config.parse(tbump_path)

    bumper = tbump.file_bumper.FileBumper(tmp_path)
    bumper.set_config(config)
    with pytest.raises(tbump.file_bumper.CurrentVersionNotFound) as e:
        bumper.compute_patches(new_version="1.3.1")
    assert e.value.src == "version.txt"
    e.value.print_error()


def test_replacing_with_empty_groups(tmp_path):
    tbump_path = tmp_path.joinpath("tbump.toml")
    tbump_path.write_text(
        r"""
        [version]
        current = "1.2.3"
        regex = '''
            (?P<major>\d+)
            \.
            (?P<minor>\d+)
            (
              \.
              (?P<patch>\d+)
            )?
        '''

        [git]
        message_template = "Bump to {new_version}"
        tag_template = "v{new_version}"

        [[file]]
        src = "foo"
        version_template = "{major}.{minor}.{patch}"

        """
    )
    foo_path = tmp_path.joinpath("foo")
    foo_path.write_text(
        """
        version = "1.2.3"
        """
    )

    bumper = tbump.file_bumper.FileBumper(tmp_path)
    config = tbump.config.parse(tbump_path)
    bumper.set_config(config)
    with pytest.raises(tbump.file_bumper.BadSubstitution) as e:
        bumper.compute_patches(new_version="1.3")
    assert e.value.groups == {"major": "1", "minor": "3", "patch": None}


def test_changing_same_file_twice(tmp_path):
    tbump_path = tmp_path.joinpath("tbump.toml")
    tbump_path.write_text(
        r"""
        [version]
        current = "1.2.3"
        regex = '''
            (?P<major>\d+)
            \.
            (?P<minor>\d+)
            (
              \.
              (?P<patch>\d+)
            )?
        '''

        [git]
        message_template = "Bump to {new_version}"
        tag_template = "v{new_version}"

        [[file]]
        src = "foo.c"
        version_template = "{major}.{minor}"
        search = "PUBLIC_VERSION"

        [[file]]
        src = "foo.c"
        search = "FULL_VERSION"

        """
    )

    foo_c = tmp_path.joinpath("foo.c")
    foo_c.write_text(
        """
        #define FULL_VERSION "1.2.3"
        #define PUBLIC_VERSION "1.2"
        """
    )
    bumper = tbump.file_bumper.FileBumper(tmp_path)
    config = tbump.config.parse(tbump_path)
    bumper.set_config(config)
    patches = bumper.compute_patches(new_version="1.3.0")
    bumper.apply_patches(patches)

    assert_in_file(foo_c, '#define FULL_VERSION "1.3.0"')
    assert_in_file(foo_c, '#define PUBLIC_VERSION "1.3"')



import toml
from tbump.test.conftest import assert_in_file

import pytest

import tbump.hooks
import tbump.main
import tbump.git


def check_file_contents(test_repo):
    toml_path = test_repo.joinpath("tbump.toml")
    new_toml = toml.loads(toml_path.text())
    assert new_toml["version"]["current"] == "1.2.41-alpha-2"

    assert_in_file("package.json", '"version": "1.2.41-alpha-2"')
    assert_in_file("package.json", '"other-dep": "1.2.41-alpha-1"')
    assert_in_file("pub.js", "PUBLIC_VERSION = '1.2.41'")


def check_git_status(test_repo):
    check_local_git_status(test_repo)


def check_local_git_status(test_repo):
    _, out = tbump.git.run_git_captured(test_repo, "log", "--oneline")
    assert "Bump to 1.2.41-alpha-2" in out

    _, out = tbump.git.run_git_captured(test_repo, "tag", "--list")
    assert "v1.2.41-alpha-2" in out


def has_pushed(test_repo, pushed=True):
    tbump.git.run_git(test_repo, "ls-remote", "origin", "refs/tags/v1.2.41-alpha-2")

    _, local_commit = tbump.git.run_git_captured(test_repo, "rev-parse", "HEAD")
    _, out = tbump.git.run_git_captured(test_repo, "ls-remote", "origin", "refs/heads/master")
    remote_commit = out.split()[0]
    return remote_commit == local_commit


def test_end_to_end(test_repo):
    tbump.main.main(["-C", test_repo, "1.2.41-alpha-2", "--non-interactive"])

    check_file_contents(test_repo)
    check_git_status(test_repo)
    assert has_pushed(test_repo)


def test_tbump_toml_not_found(test_repo, message_recorder):
    toml_path = test_repo.joinpath("tbump.toml")
    toml_path.remove()
    with pytest.raises(SystemExit):
        tbump.main.main(["-C", test_repo, "1.2.42", "--non-interactive"])
    assert message_recorder.find("No such file")


def test_tbump_toml_bad_syntax(test_repo, message_recorder):
    toml_path = test_repo.joinpath("tbump.toml")
    bad_toml = toml.loads(toml_path.text())
    del bad_toml["git"]
    toml_path.write_text(toml.dumps(bad_toml))
    with pytest.raises(SystemExit):
        tbump.main.main(["-C", test_repo, "1.2.42", "--non-interactive"])
    assert message_recorder.find("Missing keys")


def test_new_version_does_not_match(test_repo, message_recorder):
    with pytest.raises(SystemExit):
        tbump.main.main(["-C", test_repo, "1.2.41a2", "--non-interactive"])
    assert message_recorder.find("Could not parse 1.2.41a2")


def test_abort_if_file_does_not_exist(test_repo, message_recorder):
    test_repo.joinpath("package.json").remove()
    tbump.git.run_git(test_repo, "add", "--update")
    tbump.git.run_git(test_repo, "commit", "--message", "remove package.json")
    with pytest.raises(SystemExit):
        tbump.main.main(["-C", test_repo, "1.2.41-alpha-2", "--non-interactive"])
    assert message_recorder.find("package.json does not exist")


def test_interactive_abort(test_repo, mock):
    ask_mock = mock.patch("ui.ask_yes_no")
    ask_mock.return_value = False

    with pytest.raises(SystemExit):
        tbump.main.main(["-C", test_repo, "1.2.41-alpha-2"])

    ask_mock.assert_called_with("Looking good?", default=False)
    assert_in_file("VERSION", "1.2.41-alpha-1")
    rc, out = tbump.git.run_git_captured(test_repo, "tag", "--list")
    assert "v1.2.42-alpha-2" not in out


def test_abort_if_dirty(test_repo, message_recorder):
    test_repo.joinpath("VERSION").write_text("unstaged changes\n", append=True)

    with pytest.raises(SystemExit):
        tbump.main.main(["-C", test_repo, "1.2.41-alpha-2", "--non-interactive"])
    assert message_recorder.find("dirty")


def test_abort_if_tag_exists(test_repo, message_recorder):
    tbump.git.run_git(test_repo, "tag", "v1.2.42")

    with pytest.raises(SystemExit):
        tbump.main.main(["-C", test_repo, "1.2.42", "--non-interactive"])
    assert message_recorder.find("1.2.42 already exists")


def test_abort_if_file_does_not_match(test_repo, message_recorder):
    invalid_src = test_repo.joinpath("foo.txt")
    invalid_src.write_text("this is foo")
    tbump_path = test_repo.joinpath("tbump.toml")
    tbump_path.write_text("""\
    [[file]]
    src = "foo.txt"
    """, append=True)
    tbump.git.run_git(test_repo, "add", ".")
    tbump.git.run_git(test_repo, "commit", "--message", "add foo.txt")

    with pytest.raises(SystemExit):
        tbump.main.main(["-C", test_repo, "1.2.42", "--non-interactive"])
    assert message_recorder.find("not found")
    assert_in_file("VERSION", "1.2.41-alpha-1")


def test_no_tracked_branch_proceed_and_skip_push(test_repo, mock):
    ask_mock = mock.patch("ui.ask_yes_no")

    ask_mock.return_value = [True, True]
    tbump.git.run_git(test_repo, "checkout", "-b", "devel")

    tbump.main.main(["-C", test_repo, "1.2.42"])

    ask_mock.assert_has_calls(
        ask_mock.call("Continue anyway?", default=False),
        ask_mock.call("Looking good?", default=False),
    )
    assert_in_file("VERSION", "1.2.42")


def test_no_tracked_branch_but_ref_exists(test_repo, mock, message_recorder):
    ask_mock = mock.patch("ui.ask_yes_no")

    ask_mock.return_value = [True]
    tbump.git.run_git(test_repo, "checkout", "-b", "devel")

    with pytest.raises(SystemExit):
        tbump.main.main(["-C", test_repo, "1.2.41-alpha-1"])
    assert message_recorder.find("already exists")


def test_no_tracked_branch_cancel(test_repo, mock, message_recorder):
    ask_mock = mock.patch("ui.ask_yes_no")
    ask_mock.return_value = False
    tbump.git.run_git(test_repo, "checkout", "-b", "devel")

    with pytest.raises(SystemExit):
        tbump.main.main(["-C", test_repo, "1.2.42"])

    ask_mock.assert_called_with("Continue anyway?", default=False)
    assert_in_file("VERSION", "1.2.41")
    assert message_recorder.find("Cancelled")


def test_no_tracked_branch_non_interactive(test_repo,  message_recorder):
    tbump.git.run_git(test_repo, "checkout", "-b", "devel")

    with pytest.raises(SystemExit):
        tbump.main.main(["-C", test_repo, "1.2.42", "--non-interactive"])
    assert message_recorder.find("Cannot push")


def test_interactive_proceed(test_repo, mock):
    ask_mock = mock.patch("ui.ask_yes_no")

    ask_mock.return_value = [True]
    tbump.main.main(["-C", test_repo, "1.2.42"])
    ask_mock.assert_called_with("Looking good?", default=False)
    _, out = tbump.git.run_git_captured(test_repo, "ls-remote")
    assert "tags/v1.2.42" in out


def test_do_not_add_untracked_files(test_repo):
    test_repo.joinpath("untracked.txt").write_text("please don't add me")
    tbump.main.main(["-C", test_repo, "1.2.42", "--non-interactive"])
    _, out = tbump.git.run_git_captured(test_repo, "show", "--stat", "HEAD")
    assert "untracked.txt" not in out


def test_bad_substitution(test_repo, message_recorder):
    toml_path = test_repo.joinpath("tbump.toml")
    new_toml = toml.loads(toml_path.text())
    new_toml["file"][0]["version_template"] = "{release}"
    toml_path.write_text(toml.dumps(new_toml))
    tbump.git.run_git(test_repo, "add", ".")
    tbump.git.run_git(test_repo, "commit", "--message", "update repo")
    with pytest.raises(SystemExit):
        tbump.main.main(["-C", test_repo, "1.2.42"])
    assert message_recorder.find("refusing to replace by version containing 'None'")
